@@ -22,6 +22,7 @@
  */
 
 #include "config.h"
+#include "queue.h"
 
 #include <errno.h>
 #include <string.h>
@@ -295,7 +296,7 @@ static int mroute4_del_vif(struct iface *iface)
 	int rc = 0;
 
 	if (iface->mrdisc)
-		mrdisc_deregister(iface->vif);
+		rc = mrdisc_deregister(iface->vif);
 
 	if (kern_vif_del(iface)) {
 		switch (errno) {
@@ -309,9 +310,9 @@ static int mroute4_del_vif(struct iface *iface)
 		rc = -1;
 	}
 
-	if (iface->vif >= 0 && iface->vif < ALL_VIFS)
+	if (iface->vif != ALL_VIFS)
 		mroute4_prune_vif(iface->vif);
-	iface->vif = -1;
+	iface->vif = ALL_VIFS;
 
 	return rc;
 }
@@ -534,13 +535,23 @@ static int mfc_install(struct mroute *route)
 	return 0;
 }
 
+/*
+ * When route has an empty oif list -- attempt full removal of the
+ * route, unless there exist other configured routes that map to the
+ * same kernel MFC entry.
+ *
+ * When route oif list is *not* empty, attempt to remove only select
+ * interfaces from the MFC entry.  Again, unless other configured
+ * routes map to the same MFC entry.
+ */
 static int mfc_uninstall(struct mroute *route)
 {
-	struct mroute *conf, *kern;
+	struct mroute *conf, *kern, *tmp;
+	int removal = !is_active(route);
 	int diff = 0;
 	int rc = 0;
 
-	TAILQ_FOREACH(kern, &kern_list, link) {
+	TAILQ_FOREACH_SAFE(kern, &kern_list, link, tmp) {
 		if (!is_match(route, kern))
 			continue;
 
@@ -549,7 +560,7 @@ static int mfc_uninstall(struct mroute *route)
 
 		/* First remove OIFs from route entry */
 		for (size_t i = 0; i < NELEMS(route->ttl); i++) {
-			if (route->ttl[i] > 0) {
+			if (removal || route->ttl[i] > 0) {
 				kern->ttl[i] = 0;
 				diff++;
 			}
@@ -568,10 +579,10 @@ static int mfc_uninstall(struct mroute *route)
 			}
 		}
 
-		if (!diff && is_active(route))
+		if (!diff && !removal)
 			continue;
 
-		if (is_active(kern) || is_active(route)) {
+		if (is_active(kern) || !removal) {
 			rc += kern_mroute_add(kern);
 			continue;
 		}
@@ -668,7 +679,7 @@ int mroute_del_route(struct mroute *route)
 	} else {
 	cleanup:
 		TAILQ_REMOVE(&conf_list, conf, link);
-		rc = mfc_uninstall(conf);
+		rc = mfc_uninstall(route);
 		free(conf);
 	}
 
@@ -911,9 +922,9 @@ static int mroute6_del_mif(struct iface *iface)
 		rc = -1;
 	}
 
-	if (iface->mif >= 0 && iface->mif < ALL_VIFS)
+	if (iface->mif != ALL_VIFS)
 		mroute6_prune_mif(iface->mif);
-	iface->mif = -1;
+	iface->mif = ALL_VIFS;
 
 	return rc;
 }
